# 1 "../BM13_Energy.st"
# 1 "<built-in>"
# 1 "<command-line>"
# 1 "../BM13_Energy.st"

program BM13_Energy("E=XXBMA:E, MONO=XXBMA:m17, EXPTAB_Z=XXBMD:m22, YXTAL=XXBMA:MON:, ZXTAL=XXBMA:m14")
# 33 "../BM13_Energy.st"
%%#include <math.h>
%%#include <epicsThread.h>


double en; assign en to "{E}:Energy.VAL"; monitor en;
double en_rbv; assign en_rbv to "{E}:E_RBV";
double dspace; assign dspace to "{E}:dspace.VAL"; monitor dspace;
double height; assign height to "{E}:height.VAL"; monitor height;
double width; assign width to "{E}:width.VAL"; monitor width;
int verbose; assign verbose to "{E}:verbose.VAL"; monitor verbose;
int en_busy; assign en_busy to "{E}:Moving.VAL"; monitor en_busy;
int en_set; assign en_set to "{E}:SET.VAL"; monitor en_set;






int y2_track; assign y2_track to "{E}:y2_track.VAL"; monitor y2_track;
int z2_track; assign z2_track to "{E}:z2_track.VAL"; monitor z2_track;
int xt_track; assign xt_track to "{E}:xt_track.VAL"; monitor xt_track;

double xt_off; assign xt_off to "{E}:xt_off.VAL"; monitor xt_off;
double xt_sign; assign xt_sign to "{E}:xt_sign.VAL"; monitor xt_sign;
double xt_mvmin; assign xt_mvmin to "{E}:xt_mvmin.VAL"; monitor xt_mvmin;
double xt_ht; assign xt_ht to "{EXPTAB_Z}.VAL"; monitor xt_ht;
int xt_done; assign xt_done to "{EXPTAB_Z}.DMOV"; monitor xt_done;


double z2; assign z2 to "{ZXTAL}.VAL"; monitor z2;
double z2lo; assign z2lo to "{ZXTAL}.LLM"; monitor z2lo;
double z2hi; assign z2hi to "{ZXTAL}.HLM"; monitor z2hi;
int z2_done; assign z2_done to "{ZXTAL}.DMOV"; monitor z2_done;
double y2; assign y2 to "{YXTAL}t1.Y"; monitor y2;
int y2_done; assign y2_done to "{YXTAL}T1:done.VAL"; monitor y2_done;


double th; assign th to "{MONO}"; monitor th;
double th_rbv; assign th_rbv to "{MONO}.RBV"; monitor th_rbv;
int th_set; assign th_set to "{MONO}.SET";
int th_done; assign th_done to "{MONO}.DMOV"; monitor th_done;


evflag ef_en; sync en ef_en;
evflag ef_y2_track; sync y2_track ef_y2_track;
evflag ef_xt_track; sync xt_track ef_xt_track;
evflag ef_width; sync width ef_width;
evflag ef_dspace; sync dspace ef_dspace;
evflag ef_th; sync th ef_th;
evflag ef_th_rbv; sync th_rbv ef_th_rbv;


double d2r;
double r2d;
double mil;
double hc;
double omega;
double en_min;
double en_max;
double th_min;
double th_max;
double xt_diff;
double xt_old;
double xt_new;
float d_time;
double tiny;
int still_busy;
int wait_count;

ss Energy {
  state init {
    when() {
      printf("   Energy > Energy State Set initializing \n");
      en_busy = 1; pvPut(en_busy);

      if ((dspace < 0.001) || (dspace > 1000.)) {
        dspace = 1.63749;
        pvPut(dspace);
      }
      r2d = 57.29577951; d2r = 1./r2d; mil = 0.001;
      hc = 12398.61;
      omega = hc/(2 * dspace);
      en_min = 4125.000 ;
      en_max = 100 * omega ;
      th_min = r2d * asin(omega/en_max);
      th_max = r2d * asin(omega/en_min);
      tiny = 1.e-5;


      if (th < th_min ) th = th_min;
      en = omega / sin(th * d2r);
      if (th_rbv < th_min ) th_rbv = th_min;
      en_rbv = omega / sin(th_rbv * d2r) ;
      en = en_rbv;
      pvPut(en); pvPut(en_rbv);
      if (verbose) { printf("   E(init)> mono angle, energy = %12.6f %12.6f\n", th_rbv,en_rbv); }

      xt_off = xt_ht - xt_sign * 2 * width * cos(th * d2r);
      pvPut(xt_off);

      efClear(ef_en);
      en_busy = 0; pvPut(en_busy);
      if (verbose) { printf("   E(init)> done initializing\n");}
      efClear(ef_th);
    } state monitor_changes
  }


  state monitor_changes {
    when(efTestAndClear(ef_en)) {
# 157 "../BM13_Energy.st"
      en_busy = 1; pvPut(en_busy);

      if (en < 0.0) en = -en;
      if (en < 500.0) en = en * 1000.;
      if (en < en_min) en = en_min;
      if (en > en_max) en = en_max;


      pvPut(en); efClear(ef_en);
      if (en_set) {
        if (verbose) { printf("   E(set)> mono energy -> %12.6f\n", en);}
        pvPut(en);
        th = asin(omega / en) * r2d;
        th_set = 1; pvPut(th_set);
        pvPut(th);
        th_set = 0; pvPut(th_set);
        if (verbose) { printf("   E(set)> mono angle  -> %12.6f\n", th);}
      } else {
        if (verbose) { printf("  E(move)> mono energy -> %12.6f\n", en);}



        th = (asin(omega / en) * r2d) - tiny;
        if (th < th_min) th = th_min;
        if (verbose) { printf("  E(move)>    mono angle -> %12.6f \n", th);}
        pvPut(th);

        if (z2_track) {
          z2 = height * en / (2 * omega);
          if (z2 < z2lo) z2 = z2lo + tiny;
          if (z2 > z2hi) z2 = z2hi - tiny;
          pvPut(z2);
          if (verbose) { printf("  E(move)>    2nd Xtal Z -> %12.6f\n",z2);}
        }

        if (y2_track) {
          width = height / (2.0 * cos(th * d2r));
          pvPut(width);
          y2 = width;
          pvPut(y2);
          if (verbose) { printf("  E(move)> 2nd XTAL y2,width,height -> %12.6f %12.6f %12.6f\n",y2, width, height);}
        } else {




        }

        if (xt_track) {
          xt_old = xt_ht;
   pvGet(width);
          xt_new = xt_off + 2.00 * xt_sign * width * cos(th * d2r);;
          xt_diff = (xt_old *10000.0 - xt_new * 10000.0 ) / 10000.0 ;
          if (xt_diff < 0.0) xt_diff = - xt_diff;
   if (verbose) { printf("  E(move)> exptab from %12.6f to %12.6f \n",xt_old,xt_new);}
          if (xt_diff > xt_mvmin) {
            xt_ht = xt_new;
            pvPut (xt_ht);
            if (verbose) { printf("  E(move)>    XAS Table Z -> %12.6f\n",xt_ht);}
          }
        }





        d_time = 0.1;
        for (wait_count=0, still_busy=1; (wait_count<256) && still_busy;
                                                                    wait_count++) {
          epicsThreadSleep(d_time);
          still_busy = ((y2_track && (y2_done == 0)) ||
                        (z2_track && (z2_done == 0)) ||
                        (xt_track && (xt_done == 0)) ||
                        (th_done == 0));
          if (verbose) {
            printf("[%i%i%i%i]", th_done, y2_done, z2_done, xt_done);
            if (!(wait_count % 8)) printf("\n");
          }
        }
        if ((verbose) && (wait_count % 8)) {printf("\n");}
      }
      en_busy = 0;
      pvPut(en_busy);
      efClear(ef_th);
      efClear(ef_width);
      if (verbose) {printf("  E(move)> done, mono theta = %12.6f, E = %12.6f\n",
                           th_rbv,en_rbv); }
    } state monitor_changes
# 255 "../BM13_Energy.st"
    when( efTestAndClear(ef_th) ) {
      if (verbose) {printf("  E(mono)> Mono angle changed: %12.6f\n", th); }
      if (th < th_min) {
        th = th_min;
      }
      en = omega / sin(th * d2r) ;
      pvPut(en);
      if (verbose) {printf("  E(mono)> Energy being set to %12.6f %12.6f\n", en, th);}

      pvGet(th_rbv);
      if (th_rbv < th_min ) th_rbv = th_min;
      en_rbv = omega / sin(th_rbv * d2r) ;
      if (verbose) {printf("  E(mono)> Energy rbv = %12.6f %12.6f\n", en_rbv, th_rbv);}
      pvPut(en_rbv);
      efClear(ef_en);
      efClear(ef_th);
    } state monitor_changes

    when( efTestAndClear(ef_th_rbv) ) {
      if (th_rbv < 0.001 ) th_rbv = 0.001;
      en_rbv = omega / sin(th_rbv * d2r) ;
      pvPut(en_rbv);
    } state monitor_changes


    when( efTestAndClear(ef_xt_track) ) {
      if (xt_track) {
        xt_off = xt_ht - 2.00 * width * xt_sign * cos(th * d2r);
        pvPut(xt_off);
        if (verbose) {printf("  E(xt_track)> changed table offset => %12.6f\n",xt_off);}
      }
    } state monitor_changes
# 296 "../BM13_Energy.st"
   when( efTestAndClear(ef_dspace)) {
      if (dspace < 0.0) dspace = -dspace;
      if (dspace > 100.0) dspace = dspace / 1000.;
      if (dspace < 0.01 ) dspace = dspace * 1000.;
      omega = hc/(2 * dspace);
      en_min = 3000.000 ;
      en_max = 100 * omega ;
      th_min = r2d * asin(omega/en_max);
      th_max = r2d * asin(omega/en_min);
      pvPut(dspace);
      if (verbose) {printf("  E> dspace => %12.6f\n",dspace);}
    } state monitor_changes


   when( efTestAndClear(ef_width)) {
      if (width < 0.0) width = -width;
      if (width >80.0) width = 80.0;
      if (width < 5. ) width = 5.0;
      pvPut(width);
      if (verbose) {printf("  E> width => %12.6f\n",width);}
    } state monitor_changes
  }
}
